# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module to display deprecation warnings."""

import warnings


def deprecate(message: str, level: int) -> None:
    """Issues a deprecation warning using the FutureWarning category.

    Args:
        message (str): The deprecation warning message.
        level (int): The `stacklevel`, can be used to refer to `deprecate()`'s
            caller instead of the source of `deprecate()` itself.

    """
    warnings.warn(message, category=FutureWarning, stacklevel=level)
