# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Abstract Node class for all nodes.
"""


# pylint: disable=unused-import
from peekingduck.pipeline.nodes.abstract_node import AbstractNode
from peekingduck.utils.deprecation import deprecate

deprecate(
    "importing AbstractNode from peekingduck.pipeline.nodes.node is deprecated "
    "and will be removed in a future version. Please import from "
    "peekingduck.pipeline.nodes.abstract_node instead.",
    3,
)
